import typing

import attr
import qts
from qts import QtCore
from qts import QtGui
from qts import QtWidgets
import qtrio
import trio
import trio_typing


class QSignaledWidget(QtWidgets.QWidget):
    """A :class:`QtWidgets.QWidget` with extra signals for events of interest.

    Attributes:
        closed: A signal that will be emitted after a close event.
    """

    closed = QtCore.Signal()
    shown = QtCore.Signal()

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """Detect close events and emit the ``closed`` signal."""

        super().closeEvent(event)
        if event.isAccepted():
            # TODO: https://bugreports.qt.io/browse/PYSIDE-1318
            if qts.is_pyqt_5_wrapper:
                self.closed.emit()
            elif qts.is_pyside_5_wrapper:
                signal = typing.cast(QtCore.SignalInstance, self.closed)
                signal.emit()
            else:  # pragma: no cover
                raise qtrio.InternalError(
                    "You should not be here but you are running neither PyQt5 nor PySide2.",
                )
        else:  # pragma: no cover
            pass

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        """Detect show events and emit the ``shown`` signal."""

        super().showEvent(event)
        if event.isAccepted():
            # TODO: https://bugreports.qt.io/browse/PYSIDE-1318
            if qts.is_pyqt_5_wrapper:
                self.shown.emit()
            elif qts.is_pyside_5_wrapper:
                signal = typing.cast(QtCore.SignalInstance, self.shown)
                signal.emit()
            else:  # pragma: no cover
                raise qtrio.InternalError(
                    "You should not be here but you are running neither PyQt5 nor PySide2.",
                )
        else:  # pragma: no cover
            pass


@attr.s(auto_attribs=True)
class Widget:
    """A manager for a simple window with increment and decrement buttons to change a
    counter which is displayed via a widget in the center.
    """

    widget: QSignaledWidget = attr.ib(factory=QSignaledWidget)
    increment: QtWidgets.QPushButton = attr.ib(factory=QtWidgets.QPushButton)
    decrement: QtWidgets.QPushButton = attr.ib(factory=QtWidgets.QPushButton)
    label: QtWidgets.QLabel = attr.ib(factory=QtWidgets.QLabel)
    layout: QtWidgets.QHBoxLayout = attr.ib(factory=QtWidgets.QHBoxLayout)
    count: int = 0
    serving_event: trio.Event = attr.ib(factory=trio.Event)

    def setup(self, title: str, parent: typing.Optional[QtWidgets.QWidget]) -> None:
        self.widget.setParent(parent)

        self.widget.setWindowTitle(title)
        self.widget.setLayout(self.layout)

        self.increment.setText("+")
        self.decrement.setText("-")

        self.label.setText(str(self.count))

        self.layout.addWidget(self.decrement)
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.increment)

    def increment_count(self) -> None:
        """Increment the counter and update the label."""

        self.count += 1
        self.label.setText(str(self.count))

    def decrement_count(self) -> None:
        """Decrement the counter and update the label."""

        self.count -= 1
        self.label.setText(str(self.count))

    async def show(self) -> None:
        """Show the primary widget for this window."""

        self.widget.show()

    async def serve(
        self,
        *,
        task_status: trio_typing.TaskStatus[None] = trio.TASK_STATUS_IGNORED,
    ) -> None:
        signals = [
            self.decrement.clicked,
            self.increment.clicked,
            self.widget.closed,
        ]

        async with qtrio.enter_emissions_channel(signals=signals) as emissions:
            await self.show()
            task_status.started()
            self.serving_event.set()

            async for emission in emissions.channel:
                if emission.is_from(self.decrement.clicked):
                    self.decrement_count()
                elif emission.is_from(self.increment.clicked):
                    self.increment_count()
                elif emission.is_from(self.widget.closed):
                    break
                else:  # pragma: no cover
                    raise qtrio.QTrioException(f"Unexpected emission: {emission}")


async def start_widget(
    title: str = "QTrio Emissions Example",
    parent: typing.Optional[QtWidgets.QWidget] = None,
    hold_event: typing.Optional[trio.Event] = None,
    *,
    cls: typing.Type[Widget] = Widget,
    task_status: trio_typing.TaskStatus[Widget] = trio.TASK_STATUS_IGNORED,
) -> None:
    self = cls()
    self.setup(title=title, parent=parent)

    task_status.started(self)

    if hold_event is not None:
        await hold_event.wait()

    await self.serve()



import trio

import qtrio.examples.readme.qtrio_example


async def test_main():
    text_to_enter = "everyone"

    async with trio.open_nursery() as nursery:
        dialogs = await nursery.start(qtrio.examples.readme.qtrio_example.main)

        assert dialogs.input.line_edit is not None
        dialogs.input.line_edit.setText(text_to_enter)

        assert dialogs.input.accept_button is not None
        async with qtrio._core.wait_signal_context(dialogs.output.shown):
            dialogs.input.accept_button.click()

        assert dialogs.output.dialog is not None
        output_text = dialogs.output.dialog.text()

        assert dialogs.output.accept_button is not None
        dialogs.output.accept_button.click()

        assert text_to_enter in output_text


async def test_main_cancelled():
    text_to_enter = "everyone"

    def output_was_shown(*args, **kwargs):
        assert False, "Output dialog was shown"  # pragma: no cover

    async with trio.open_nursery() as nursery:
        dialogs = await nursery.start(async_fn=qtrio.examples.readme.qtrio_example.main)

        async with qtrio.open_emissions_nursery() as emissions_nursery:
            emissions_nursery.connect_sync(
                signal=dialogs.output.shown,
                slot=output_was_shown,
            )
            assert dialogs.output.dialog is None

            assert dialogs.input.line_edit is not None
            dialogs.input.line_edit.setText(text_to_enter)

            assert dialogs.input.reject_button is not None
            dialogs.input.reject_button.click()



"""Thanks to njsmith for developing Trio guest mode and sharing the first example
integration with Qt.
"""
import contextlib
import functools
import math
import os
import time
import typing

import attr
import httpcore._async.http11
import httpx
import hyperlink
import trio
import trio_typing

import qtrio
import qtrio.dialogs


# Default is 4096
httpcore._async.http11.AsyncHTTP11Connection.READ_NUM_BYTES = 100_000

default_fps: int = 60


def create_title(specific: str) -> str:
    return f"QTrio Download Example - {specific}"


@attr.s(auto_attribs=True, frozen=True)
class Progress:
    downloaded: int
    first: bool
    total: typing.Optional[int] = None


@attr.s(auto_attribs=True, eq=False)
class Downloader:
    text_input_dialog: typing.Optional[qtrio.dialogs.TextInputDialog] = None
    file_dialog: typing.Optional[qtrio.dialogs.FileDialog] = None
    get_dialog: typing.Optional["GetDialog"] = None

    text_input_shown_event: trio.Event = attr.ib(factory=trio.Event)
    file_dialog_shown_event: trio.Event = attr.ib(factory=trio.Event)
    get_dialog_created_event: trio.Event = attr.ib(factory=trio.Event)

    async def serve(
        self,
        url: typing.Optional[typing.Union[str, hyperlink.URL]] = None,
        destination: typing.Optional[typing.Union[str, os.PathLike]] = None,
        fps: float = default_fps,
        http_application: typing.Optional[typing.Callable[..., typing.Any]] = None,
        *,
        task_status: trio_typing.TaskStatus[None] = trio.TASK_STATUS_IGNORED,
    ) -> None:
        task_status.started()

        converted_url: hyperlink.URL
        converted_destination: trio.Path

        with contextlib.suppress(qtrio.UserCancelledError):
            if url is None:
                # async with trio.open_nursery() as nursery:
                #     start = functools.partial(
                #         qtrio.dialogs.TextInputDialog.serve,
                #         title=create_title("Enter URL"),
                #         label="URL to download:",
                #     )
                #     self.text_input_dialog = await nursery.start(start)

                self.text_input_dialog = qtrio.dialogs.create_text_input_dialog()

                self.text_input_dialog.title = create_title("Enter URL")
                self.text_input_dialog.label = "URL to download:"

                url = await self.text_input_dialog.wait(
                    shown_event=self.text_input_shown_event,
                )
                self.text_input_dialog = None

            if isinstance(url, str):
                converted_url = hyperlink.URL.from_text(url)
            else:
                converted_url = url

            if destination is None:
                self.file_dialog = qtrio.dialogs.create_file_save_dialog()

                default_file: str = ""
                if len(converted_url.path) > 0:
                    default_file = converted_url.path[-1]

                self.file_dialog.default_file = trio.Path(default_file)

                destination = await self.file_dialog.wait(
                    shown_event=self.file_dialog_shown_event,
                )
                self.file_dialog = None

            converted_destination = trio.Path(destination)

            async with trio.open_nursery() as nursery:
                start = functools.partial(
                    start_get_dialog,
                    url=converted_url,
                    destination=converted_destination,
                    fps=fps,
                    http_application=http_application,
                )
                self.get_dialog = await nursery.start(start)
                self.get_dialog_created_event.set()


async def start_downloader(
    url: typing.Optional[hyperlink.URL] = None,
    destination: typing.Optional[trio.Path] = None,
    fps: float = default_fps,
    http_application: typing.Optional[typing.Callable[..., typing.Any]] = None,
    hold_event: typing.Optional[trio.Event] = None,
    *,
    cls: typing.Type[Downloader] = Downloader,
    task_status: trio_typing.TaskStatus[Downloader] = trio.TASK_STATUS_IGNORED,
) -> None:
    self = cls()

    task_status.started(self)

    if hold_event is not None:
        await hold_event.wait()

    await self.serve(
        url=url, destination=destination, fps=fps, http_application=http_application
    )


@attr.s(auto_attribs=True, eq=False)
class GetDialog:
    fps: float = default_fps
    clock: typing.Callable[[], float] = time.monotonic
    http_application: typing.Optional[typing.Callable[..., typing.Any]] = None

    progress_dialog: typing.Optional[qtrio.dialogs.ProgressDialog] = None
    message_box: typing.Optional[qtrio.dialogs.MessageBox] = None

    progress_dialog_shown_event: trio.Event = attr.ib(factory=trio.Event)
    message_box_shown_event: trio.Event = attr.ib(factory=trio.Event)

    async def serve(
        self,
        url: hyperlink.URL,
        destination: trio.Path,
    ) -> None:
        self.progress_dialog = qtrio.dialogs.create_progress_dialog()

        self.progress_dialog.title = create_title("Fetching")
        self.progress_dialog.text = f"Fetching {url}..."

        async with self.progress_dialog.manage():
            if self.progress_dialog.dialog is None:  # pragma: no cover
                raise qtrio.InternalError(
                    "Dialog not assigned while it is being managed."
                )

            # Always show the dialog
            self.progress_dialog.dialog.setMinimumDuration(0)
            self.progress_dialog_shown_event.set()

            start = self.clock()

            async for progress in get(
                url=url,
                destination=destination,
                update_period=1 / self.fps,
                clock=self.clock,
                http_application=self.http_application,
            ):
                if progress.first:
                    if progress.total is None:
                        maximum = 0
                    else:
                        maximum = progress.total

                    self.progress_dialog.dialog.setMaximum(maximum)
                    self.progress_dialog.dialog.setValue(0)

                if progress.total is not None:
                    self.progress_dialog.dialog.setValue(progress.downloaded)

            end = self.clock()

        self.progress_dialog = None

        duration = end - start
        if duration == 0:
            # define this seems to happen when testing on Windows with an x86 Python
            if progress.downloaded > 0:
                bytes_per_second = math.inf
            else:  # pragma: no cover
                bytes_per_second = 0
        else:
            bytes_per_second = progress.downloaded / duration

        summary = "\n\n".join(
            [
                url.asText(),
                os.fspath(destination),
                f"Downloaded {progress.downloaded} bytes in {duration:.2f} seconds",
                f"{bytes_per_second:.2f} bytes/second",
            ]
        )

        self.message_box = qtrio.dialogs.create_message_box()

        self.message_box.title = create_title("Download Summary")
        self.message_box.text = summary

        await self.message_box.wait(shown_event=self.message_box_shown_event)

        self.message_box = None


async def start_get_dialog(
    url: hyperlink.URL,
    destination: trio.Path,
    fps: float = default_fps,
    http_application: typing.Optional[typing.Callable[..., typing.Any]] = None,
    hold_event: typing.Optional[trio.Event] = None,
    *,
    cls: typing.Type[GetDialog] = GetDialog,
    task_status: trio_typing.TaskStatus[GetDialog] = trio.TASK_STATUS_IGNORED,
) -> None:
    self = cls(fps=fps, http_application=http_application)

    task_status.started(self)

    if hold_event is not None:
        await hold_event.wait()

    await self.serve(url=url, destination=destination)


async def get(
    url: hyperlink.URL,
    destination: trio.Path,
    update_period: float = 0.2,
    clock: typing.Callable[[], float] = time.monotonic,
    http_application: typing.Optional[typing.Callable[..., typing.Any]] = None,
) -> typing.AsyncIterable[Progress]:
    async with httpx.AsyncClient(app=http_application) as client:
        async with client.stream("GET", url.asText()) as response:
            raw_content_length = response.headers.get("content-length")
            if raw_content_length is None:
                content_length = None
            else:
                content_length = int(raw_content_length)

            progress = Progress(
                downloaded=0,
                total=content_length,
                first=True,
            )

            yield progress
            last_update = clock()

            progress = attr.evolve(progress, first=False)

            downloaded = 0

            async with (await destination.open("wb")) as file:
                async for chunk in response.aiter_raw():
                    downloaded += len(chunk)
                    await file.write(chunk)

                    if clock() - last_update > update_period:
                        progress = attr.evolve(progress, downloaded=downloaded)
                        yield progress
                        last_update = clock()

            if progress.downloaded != downloaded:
                progress = attr.evolve(progress, downloaded=downloaded)
                yield progress


if __name__ == "__main__":  # pragma: no cover
    qtrio.run(start_downloader)



import functools
import typing

import pytestqt.qtbot
import trio
import trio.testing

import qtrio
import qtrio.examples.buildingrespect


async def test_main(
    qtbot: pytestqt.qtbot.QtBot,
    optional_hold_event: typing.Optional[trio.Event],
) -> None:
    message = "test world"

    async with trio.open_nursery() as nursery:
        start = functools.partial(
            qtrio.examples.buildingrespect.start_widget,
            message=message,
            hold_event=optional_hold_event,
        )
        widget: qtrio.examples.buildingrespect.Widget = await nursery.start(start)
        qtbot.addWidget(widget.widget)

        async with qtrio.enter_emissions_channel(
            signals=[widget.text_changed],
        ) as emissions:
            if optional_hold_event is not None:
                optional_hold_event.set()

            await trio.testing.wait_all_tasks_blocked(cushion=0.01)

            # lazily click the button and collect results in the memory channel rather
            # than starting concurrent tasks.

            # close the send channel so the receive channel knows when it is done
            async with emissions.send_channel:
                for _ in message:
                    widget.button.click()

                    # give Qt etc a chance to handle the clicks before closing the channel
                    await trio.testing.wait_all_tasks_blocked(cushion=0.01)

            results: typing.List[typing.Tuple[object]] = [
                emission.args async for emission in emissions.channel
            ]

            widget.button.click()

    assert results == [
        ("t",),
        ("te",),
        ("tes",),
        ("test",),
        ("test ",),
        ("test w",),
        ("test wo",),
        ("test wor",),
        ("test worl",),
        ("test world",),
    ]



import functools
import typing

import pytestqt.qtbot
import trio
import trio.testing

import qtrio
import qtrio.examples.emissions


async def test_main(
    qtbot: pytestqt.qtbot.QtBot,
    optional_hold_event: typing.Optional[trio.Event],
) -> None:
    async with trio.open_nursery() as nursery:
        start = functools.partial(
            qtrio.examples.emissions.start_widget,
            hold_event=optional_hold_event,
        )
        widget: qtrio.examples.emissions.Widget = await nursery.start(start)
        qtbot.addWidget(widget.widget)

        if optional_hold_event is not None:
            optional_hold_event.set()
        else:
            await trio.testing.wait_all_tasks_blocked(cushion=0.1)

        await widget.serving_event.wait()

        buttons = [
            widget.increment,
            widget.increment,
            widget.increment,
            widget.decrement,
            widget.decrement,
            widget.decrement,
            widget.decrement,
        ]

        results: typing.List[str] = []

        for button in buttons:
            button.click()
            await trio.testing.wait_all_tasks_blocked(cushion=0.01)
            results.append(widget.label.text())

        widget.widget.close()

    assert results == ["1", "2", "3", "2", "1", "0", "-1"]



import typing

from qts import QtCore

import qtrio._core


class ReenterEvent(QtCore.QEvent):
    """A proper ``ReenterEvent`` for reentering into the Qt host loop."""

    def __init__(self, fn: typing.Callable[[], object]):
        if qtrio._core._reenter_event_type is None:
            message = (
                "The reenter event type must be registered before creating a reenter"
                " event."
            )
            raise qtrio.InternalError(message)

        super().__init__(qtrio._core._reenter_event_type)
        self.fn = fn


class Reenter(QtCore.QObject):
    """A ``QtCore.QObject`` for handling reenter events."""

    def event(self, event: QtCore.QEvent) -> bool:
        """Qt calls this when the object receives an event."""

        try:
            reenter_event = typing.cast(Reenter, event)
            reenter_event.fn()
            return True
        except Exception as e:
            raise qtrio.InternalError("Exception while handling a reenter event") from e



import typing

import attr
import qtrio
from qts import QtWidgets
import trio
import trio_typing


@attr.s(auto_attribs=True, eq=False)
class Widget:
    message: str
    change_delay: float = 0.5
    close_delay: float = 3
    label: QtWidgets.QLabel = attr.ib(factory=QtWidgets.QLabel)

    text_changed = qtrio.Signal(str)

    done_event: trio.Event = attr.ib(factory=trio.Event)

    def setup(self) -> None:
        self.label.setText(self.message)
        self.label.setText("")

    async def show(self) -> None:
        # TODO: maybe raise if already started?
        # start big enough to fit the whole message
        self.label.setText(self.message)
        self.label.show()
        self.label.setText("")

    def set_text(self, text: str) -> None:
        self.label.setText(text)
        self.text_changed.emit(text)

    async def serve(
        self,
        *,
        task_status: trio_typing.TaskStatus[None] = trio.TASK_STATUS_IGNORED,
    ) -> None:
        await self.show()
        task_status.started()
        partial_message = ""
        for character in self.message:
            partial_message += character

            await trio.sleep(self.change_delay)
            self.set_text(partial_message)

        await trio.sleep(self.close_delay)
        self.done_event.set()


async def start_widget(
    message: str,
    change_delay: float = 0.5,
    close_delay: float = 3,
    hold_event: typing.Optional[trio.Event] = None,
    *,
    cls: typing.Type[Widget] = Widget,
    task_status: trio_typing.TaskStatus[Widget] = trio.TASK_STATUS_IGNORED,
) -> None:
    self = cls(message=message, change_delay=change_delay, close_delay=close_delay)
    self.setup()

    task_status.started(self)

    if hold_event is not None:
        await hold_event.wait()

    await self.serve()


if __name__ == "__main__":  # pragma: no cover
    qtrio.run(start_widget, "Hello world.")



def test_importing_qtrio_does_not_import_qt(testdir):
    test_file = r"""
    import sys

    def test():
        assert 'qts' not in sys.modules, "qts should not have been imported before test"

        import qtrio

        qts_qt_modules = [
            module for module in sys.modules if module.startswith('qts.Qt')
        ]

        assert qts_qt_modules == [], "qts' Qt modules should not have been imported during test"
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess("-p", "no:pytest-qt")
    result.assert_outcomes(passed=1)


def test_importing_qtrio_does_not_import_qt_wrappers(testdir):
    test_file = r"""
    import sys

    def wrapper_modules():
        return [
            module
            for module in sys.modules
            if (
                module.startswith('PyQt')
                or module.startswith('PySide')
            )
        ]

    def test():
        assert wrapper_modules() == [], "Qt wrapper modules should not have been imported prior to the test"

        import qtrio

        assert wrapper_modules() == [], "Qt wrapper modules should not have been imported during test"
    """
    testdir.makepyfile(test_file)

    result = testdir.runpytest_subprocess("-p", "no:pytest-qt")
    result.assert_outcomes(passed=1)



import typing

import attr
import qtrio
from qts import QtWidgets
import trio
import trio_typing


@attr.s(auto_attribs=True, eq=False)
class Widget:
    message: str
    widget: QtWidgets.QWidget = attr.ib(factory=QtWidgets.QWidget)
    layout: QtWidgets.QLayout = attr.ib(factory=QtWidgets.QVBoxLayout)
    button: QtWidgets.QPushButton = attr.ib(factory=QtWidgets.QPushButton)
    label: QtWidgets.QWidget = attr.ib(factory=QtWidgets.QLabel)

    text_changed = qtrio.Signal(str)

    def setup(self) -> None:
        self.button.setText("More")

        self.layout.addWidget(self.button)
        self.layout.addWidget(self.label)
        self.widget.setLayout(self.layout)

    async def show(self) -> None:
        # TODO: maybe raise if already started?
        # start big enough to fit the whole message
        self.label.setText(self.message)
        self.widget.show()
        self.label.setText("")

    def set_text(self, text: str) -> None:
        self.label.setText(text)
        self.text_changed.emit(text)

    async def serve(
        self,
        *,
        task_status: trio_typing.TaskStatus[None] = trio.TASK_STATUS_IGNORED,
    ) -> None:
        async with qtrio.enter_emissions_channel(
            signals=[self.button.clicked]
        ) as emissions:
            i = 1
            await self.show()
            task_status.started()

            async for _ in emissions.channel:  # pragma: no branch
                self.set_text(self.message[:i])
                i += 1

                if i > len(self.message):
                    break

            # wait for another click to finish
            await emissions.channel.receive()


async def start_widget(
    message: str,
    hold_event: typing.Optional[trio.Event] = None,
    *,
    cls: typing.Type[Widget] = Widget,
    task_status: trio_typing.TaskStatus[Widget] = trio.TASK_STATUS_IGNORED,
) -> None:
    self = cls(message=message)
    self.setup()

    task_status.started(self)

    if hold_event is not None:
        await hold_event.wait()

    await self.serve()


if __name__ == "__main__":  # pragma: no cover
    qtrio.run(start_widget, "Hello world.")



import contextlib

import attr
from qts import QtWidgets
import trio
import trio_typing

import qtrio
import qtrio.dialogs


def create_input() -> qtrio.dialogs.TextInputDialog:
    return qtrio.dialogs.create_text_input_dialog(
        title="Hello",
        label="Enter your name:",
    )


def create_output() -> qtrio.dialogs.MessageBox:
    return qtrio.dialogs.create_message_box(
        title="Hello",
        text="",
        icon=QtWidgets.QMessageBox.Icon.Question,
        buttons=QtWidgets.QMessageBox.Ok,
    )


@attr.s
class Dialogs:
    input: qtrio.dialogs.TextInputDialog = attr.ib(factory=create_input)
    output: qtrio.dialogs.MessageBox = attr.ib(factory=create_output)


async def main(
    *,
    task_status: trio_typing.TaskStatus[Dialogs] = trio.TASK_STATUS_IGNORED,
) -> None:
    dialogs = Dialogs()
    task_status.started(dialogs)

    with contextlib.suppress(qtrio.UserCancelledError):
        name = await dialogs.input.wait()
        dialogs.output.text = f"Hi {name}, welcome to the team!"
        await dialogs.output.wait()


if __name__ == "__main__":  # pragma: no cover
    qtrio.run(main)



import functools
import os
import pathlib
import random
import sys
import typing

import hyperlink
import pytest
import quart_trio
import trio

import qtrio.dialogs
import qtrio.examples.download


T = typing.TypeVar("T")


async def asynchronize(sequence: typing.Iterable[T]) -> typing.AsyncIterable[T]:
    for element in sequence:
        yield element


def randomly_chunked_bytes(
    random_generator: random.Random,
    chunk_size_minimum: int,
    chunk_size_maximum: int,
    chunk_count: int,
) -> typing.List[bytes]:
    return [
        bytes(
            random_generator.randrange(256)
            for _ in range(
                random_generator.randrange(chunk_size_minimum, chunk_size_maximum)
            )
        )
        for _ in range(chunk_count)
    ]


# Just using a seeded random.Random as an easy way to get consistent data without
# storing it all in the repository.
random_generator = random.Random(0)
one_chunk: typing.List[bytes] = [
    bytes(random_generator.randrange(256) for _ in range(1_000))
]
many_chunks: typing.List[bytes] = [
    *randomly_chunked_bytes(
        random_generator=random_generator,
        chunk_size_minimum=10,
        chunk_size_maximum=20,
        chunk_count=10,
    ),
    b"",
] * 2


@pytest.fixture(
    name="chunked_data",
    params=[
        many_chunks,
        one_chunk,
    ],
    ids=["Many chunks", "One chunk"],
)
def chunked_data_fixture(request):
    return request.param


@pytest.fixture(
    name="content_length_specified",
    params=[False, True],
    ids=["Unknown content length", "Known content length"],
)
def content_length_specified_fixture(request):
    return request.param


@pytest.fixture(name="content_length")
def content_length_fixture(content_length_specified, chunked_data):
    content_length: typing.Optional[int]

    if content_length_specified:
        content_length = sum(len(chunk) for chunk in chunked_data)
    else:
        content_length = None

    return content_length


@pytest.fixture(name="pass_url", params=[True, False], ids=["pass URL", "enter URL"])
def pass_url_fixture(request):
    return request.param


@pytest.fixture(
    name="pass_destination",
    params=[True, False],
    ids=["pass destination", "enter destination"],
)
def pass_destination_fixture(request):
    return request.param


urls = [hyperlink.URL.from_text(s) for s in ["http://test/", "http://test"]]


@pytest.fixture(name="url", params=urls)
def url_fixture(request):
    return request.param


@pytest.fixture(name="http_application")
def http_application_fixture(chunked_data, content_length_specified, content_length):
    application = quart_trio.QuartTrio(__name__)

    @application.route("/")
    async def root():
        headers = {}

        if content_length_specified:
            headers["content-length"] = content_length

        return asynchronize(chunked_data), 200, headers

    return application


async def test_main(
    chunked_data: typing.List[bytes],
    http_application: quart_trio.QuartTrio,
    url: hyperlink.URL,
    pass_url: bool,
    tmp_path: pathlib.Path,
    pass_destination: bool,
    optional_hold_event: typing.Optional[trio.Event],
) -> None:
    temporary_directory = trio.Path(tmp_path)

    data = b"".join(chunked_data)
    destination = temporary_directory.joinpath("file")

    async with trio.open_nursery() as nursery:
        start = functools.partial(
            qtrio.examples.download.start_downloader,
            url=url if pass_url else None,
            destination=destination if pass_destination else None,
            fps=10,
            http_application=http_application,
            hold_event=optional_hold_event,
        )
        widget: qtrio.examples.download.Downloader = await nursery.start(start)

        if optional_hold_event is not None:
            optional_hold_event.set()

        if not pass_url:
            await widget.text_input_shown_event.wait()
            assert widget.text_input_dialog is not None

            assert widget.text_input_dialog.line_edit is not None
            widget.text_input_dialog.line_edit.setText(url.to_text())

            assert widget.text_input_dialog.accept_button is not None
            widget.text_input_dialog.accept_button.click()

        if not pass_destination:
            await widget.file_dialog_shown_event.wait()
            assert widget.file_dialog is not None

            assert widget.file_dialog.dialog is not None
            await widget.file_dialog.set_path(path=destination)

            assert widget.file_dialog.accept_button is not None
            widget.file_dialog.accept_button.click()

        await widget.get_dialog_created_event.wait()
        assert widget.get_dialog is not None

        await widget.get_dialog.message_box_shown_event.wait()
        assert widget.get_dialog.message_box is not None

        assert widget.get_dialog.message_box.accept_button is not None
        widget.get_dialog.message_box.accept_button.click()

    async with await destination.open("rb") as written_file:
        written = await written_file.read()

    assert written == data


async def test_get_dialog(
    chunked_data: typing.List[bytes],
    http_application: quart_trio.QuartTrio,
    url: hyperlink.URL,
    tmp_path: pathlib.Path,
    optional_hold_event: typing.Optional[trio.Event],
) -> None:
    temporary_directory = trio.Path(tmp_path)

    data = b"".join(chunked_data)
    destination = temporary_directory.joinpath("file")

    async with trio.open_nursery() as nursery:
        start = functools.partial(
            qtrio.examples.download.start_get_dialog,
            url=url,
            destination=destination,
            fps=0.1,
            http_application=http_application,
            hold_event=optional_hold_event,
        )
        widget: qtrio.examples.download.GetDialog = await nursery.start(start)

        if optional_hold_event is not None:
            optional_hold_event.set()

        await widget.message_box_shown_event.wait()
        assert widget.message_box is not None

        assert widget.message_box.accept_button is not None
        widget.message_box.accept_button.click()

    async with await destination.open("rb") as written_file:
        written = await written_file.read()

    assert written == data


async def test_get_dialog_canceled(
    chunked_data: typing.List[bytes],
    http_application: quart_trio.QuartTrio,
    url: hyperlink.URL,
    tmp_path: pathlib.Path,
) -> None:
    temporary_directory = trio.Path(tmp_path)

    destination = temporary_directory.joinpath("file")

    with pytest.raises(qtrio.UserCancelledError):
        async with trio.open_nursery() as nursery:
            start = functools.partial(
                qtrio.examples.download.start_get_dialog,
                url=url,
                destination=destination,
                fps=0.1,
                http_application=http_application,
            )
            widget: qtrio.examples.download.GetDialog = await nursery.start(start)

            await widget.progress_dialog_shown_event.wait()
            assert widget.progress_dialog is not None

            assert widget.progress_dialog.dialog is not None
            assert widget.progress_dialog.dialog.isVisible()
            assert widget.message_box is None

            assert widget.progress_dialog.cancel_button is not None
            widget.progress_dialog.cancel_button.click()


async def test_get(
    chunked_data: typing.List[bytes],
    content_length: typing.Optional[int],
    http_application: quart_trio.QuartTrio,
    url: hyperlink.URL,
    tmp_path: pathlib.Path,
) -> None:
    temporary_directory = trio.Path(tmp_path)

    data = b"".join(chunked_data)
    destination = temporary_directory.joinpath("file")

    progresses = []

    async for progress in qtrio.examples.download.get(
        url=url,
        destination=destination,
        update_period=0,
        http_application=http_application,
    ):
        progresses.append(progress)

    async with await destination.open("rb") as written_file:
        written = await written_file.read()

    assert written == data
    assert all(progress.total == content_length for progress in progresses)
    assert [progresses[0].downloaded, progresses[-1].downloaded] == [0, len(data)]



import pytestqt.qtbot
from qts import QtWidgets

import qtrio.examples.readme.qt


def test_main(qtbot: pytestqt.qtbot.QtBot, qapp: QtWidgets.QApplication) -> None:
    input_dialog = qtrio.examples.readme.qt.create_input()
    output_dialog = qtrio.examples.readme.qt.create_output()

    qtbot.add_widget(input_dialog)
    qtbot.add_widget(output_dialog)

    text_to_enter = "everyone"

    main_object = qtrio.examples.readme.qt.Main(
        application=qapp,
        input_dialog=input_dialog,
        output_dialog=output_dialog,
    )

    with qtbot.waitExposed(input_dialog, timeout=500):
        main_object.setup()

    [line_edit] = input_dialog.findChildren(QtWidgets.QLineEdit)
    line_edit.setText(text_to_enter)

    with qtbot.waitExposed(output_dialog, timeout=500):
        input_dialog.accept()

    output_text = output_dialog.text()

    output_dialog.accept()

    assert text_to_enter in output_text


def test_main_cancelled(
    qtbot: pytestqt.qtbot.QtBot, qapp: QtWidgets.QApplication
) -> None:
    input_dialog = qtrio.examples.readme.qt.create_input()
    output_dialog = qtrio.examples.readme.qt.create_output()

    qtbot.add_widget(input_dialog)
    qtbot.add_widget(output_dialog)

    text_to_enter = "everyone"

    main_object = qtrio.examples.readme.qt.Main(
        application=qapp,
        input_dialog=input_dialog,
        output_dialog=output_dialog,
    )

    with qtbot.waitExposed(input_dialog, timeout=500):
        main_object.setup()

    [line_edit] = input_dialog.findChildren(QtWidgets.QLineEdit)
    line_edit.setText(text_to_enter)
    input_dialog.reject()

    # qtbot.wait(1_000)

    output_text = output_dialog.text()

    assert output_text == ""



import functools
import typing

import pytestqt.qtbot
import trio

import qtrio
import qtrio.examples.crossingpaths


async def test_main(
    qtbot: pytestqt.qtbot.QtBot, optional_hold_event: typing.Optional[trio.Event]
) -> None:
    message = "test world"

    async with trio.open_nursery() as nursery:
        start = functools.partial(
            qtrio.examples.crossingpaths.start_widget,
            message=message,
            change_delay=0.01,
            close_delay=0.01,
            hold_event=optional_hold_event,
        )
        widget: qtrio.examples.crossingpaths.Widget = await nursery.start(start)

        async with qtrio.enter_emissions_channel(
            signals=[widget.text_changed],
        ) as emissions:
            if optional_hold_event is not None:
                optional_hold_event.set()

            async with emissions.send_channel:
                await widget.done_event.wait()

            results = [emission.args async for emission in emissions.channel]

    assert results == [
        ("t",),
        ("te",),
        ("tes",),
        ("test",),
        ("test ",),
        ("test w",),
        ("test wo",),
        ("test wor",),
        ("test worl",),
        ("test world",),
    ]
